from .copysender import CopySender
from .ftpsender import FTPSender
from .pdlsender import PDLSender
from .emailsender import EmailSender
from .securesender import SecureSender


def get_sender_class(transfer_type):
    """Class factory for generating Sender class.

    Args:
        transfer_type: String indicating which of supported Sender classes to
            return:
            - copy CopySender
            - ftp FTPSender
            - pdl PDLSender
            - email EmailSender
            - ssh SecureSender

    Raises KeyError:
        When transfer_type is not one of the above strings.

    Returns:
        One of the above children of the Sender class.
    """
    types = {
        "copy": CopySender,
        "ftp": FTPSender,
        "pdl": PDLSender,
        "email": EmailSender,
        "ssh": SecureSender,
    }
    if transfer_type not in types:
        raise KeyError(f"{transfer_type} not a supported transfer type.")
    else:
        return types[transfer_type]



#!/usr/bin/env python

# stdlib imports
import os.path


class Sender(object):
    """
    Base class for concrete subclasses that wrap around different methods of
    transmitting/deleting files.

    Each subclass should provide class variables called:
    * `_required_properties` -- A list of names of *required* properties that
      MUST be passed in to constructor.
    * `_optional_properties` -- A list of names of *optional* properties that
      CAN be passed in to constructor.

    Each subclass should implement at least three methods:

    * `send` -- do the appropriate implementation of transferring files
    * `delete` -- do the appropriate implementation of deleting transferred
      files from remote system, if possible.
    * `cancel` -- do the appropriate implementation of sending a cancel message
      to remote system, if possible.

    delete() and cancel() may do the same thing, or nothing.
    """

    _required_properties = []
    _optional_properties = []

    def __init__(
        self,
        properties=None,
        local_files=None,
        local_directory=None,
        cancelfile=".cancel",
    ):
        """
        Create a Sender object using property settings, local files and
        directories to transfer/delete.

        Args:
            properties: Dictionary of properties that are needed for a specific
                subclass of sender.
            local_files: List of local files which should be transferred to
                remote system.
            local_directory: Local directory which should be transferred to
                remote system.
            cancelfile: For Sender subclasses that actually send a file for
                cancel() actions, this allows the user to define what that
                file is called.
        """
        self._properties = properties
        for prop in self._required_properties:
            if prop not in self._properties:
                raise KeyError(f'Required property "{prop}" not specified.')

        if local_files is not None:
            if not isinstance(local_files, list):
                raise TypeError("Input files must be a list")
            for f in local_files:
                if not os.path.isfile(f):
                    raise FileNotFoundError(f"Input file {f} could not be found")

        if local_directory is not None:
            if not os.path.isdir(local_directory):
                raise FileNotFoundError(
                    f"Input directory {local_directory} could not be found"
                )
        if local_files is not None:
            self._local_files = local_files
        else:
            self._local_files = []

        self._local_directory = None
        if local_directory is not None:
            self._local_directory = local_directory

        # set the name of the cancel file
        self._cancelfile = cancelfile

    def addProperty(self, key, value):
        self._properties[key] = value

    def addFiles(self, local_files):
        for f in local_files:
            if not os.path.isfile(f):
                raise FileNotFoundError(f"Input file {f} could not be found")
        self._local_files += local_files

    def changeDirectory(self, local_directory, directory_alias=None):
        if not os.path.isdir(local_directory):
            raise FileNotFoundError(
                f"Input directory {local_directory} could not be found"
            )
        self._local_directory = local_directory
        self._directory_alias = directory_alias

    def getAllLocalFiles(self):
        """
        Recursively find all files in local directory, create a list of each
        file with full path.

        Returns:
            List of files with full path underneath input local_directory.
        """
        allfiles = []
        if not self._local_directory:
            return allfiles
        for root, _, files in os.walk(self._local_directory):
            for fname in files:
                filename = os.path.join(root, fname)
                allfiles.append(filename)
        return allfiles

    # this is implemented in the subclasses
    def send(self):
        """

        Returns:
            Tuple containing number of files sent, and a message describing
            what was done.
        """
        pass

    # this is implemented in the subclasses
    def cancel(self):
        """

        :returns:
          A message describing what was done.
        """
        pass

    def getRequiredProperties(self):
        """Return the list of names of required properties for given subclass.

        Returns:
            List of names of required properties for given subclass.
        """
        return self._required_properties

    def getOptionalProperties(self):
        """Return the list of names of optional properties for given subclass.

        Returns:
            List of names of optional properties for given subclass.
        """
        return self._optional_properties



#!/usr/bin/env python

# stdlib imports
import datetime
import os.path
import subprocess

# local imports
from .sender import Sender

DATE_TIME_FMT = "%Y-%m-%dT%H:%M:%S.%f"


def get_command_output(cmd):
    """
    Method for calling external system command.
    Args:
        cmd: String command (e.g., 'ls -l', etc.).
    Returns:
        Three-element tuple containing a boolean indicating success or failure,
        the stdout from running the command, and stderr.
    """
    proc = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout, stderr = proc.communicate()
    retcode = proc.returncode
    if retcode == 0:
        retcode = True
    else:
        retcode = False
    return (retcode, stdout, stderr)


class PDLSender(Sender):
    """Class to invoke a PDL send command on a product.

    PDLSender uses a local installation of Product Distribution Layer (PDL)
    (https://usgs.github.io/pdl/)
    to send a file or a directory, along with desired metadata to one or more
    PDL hubs.

    Required properties:
      - java String path to Java executable on this system.
      - jarfile String path to PDL .jar file on this system.
      - privatekey String path to PDL private key file.
      - configfile String path to PDL config file (specifying remote hubs,
        ports, etc.)
      - source Source of this product (i.e., contributor of the product).
        i.e. 'us'
      - type  Product type (i.e. shakemap, losspager, etc.)
      - eventsource Source of the original event ID. i.e., ci, gcmt, etc.
      - eventsourcecode Event ID (as issued by source - i.e., 2008abcd)
      - code  eventsource + eventsourcecode

    Optional properties:
      - latitude Float latitude of event.
      - longitude Float longitude of event.
      - depth Float depth of event.
      - magnitude Float magnitude of event.
      - eventtime datetime.datetime object indicating time of origin.

    Product properties:
      Any string, int, float, or datetime.datetime object which the user would
      like to pass to PDL.
    """

    _pdlcmd_pieces = [
        "[JAVA] -jar [JARFILE] --send --status=[STATUS]",
        "--source=[SOURCE] --type=[TYPE] --code=[CODE]",
        "--eventsource=[EVENTSOURCE] --eventsourcecode=[EVENTSOURCECODE]",
        "[PRODUCT_PROPERTIES] [OPTIONAL_PROPERTIES] [CMDLINE_ARGS] ",
        "--privateKey=[PRIVATEKEY]  --configFile=[CONFIGFILE] [FILE] [DIRECTORY]",
    ]
    _pdlcmd = " ".join(_pdlcmd_pieces)

    _required_properties = [
        "java",
        "jarfile",
        "privatekey",
        "configfile",
        "source",
        "type",
        "code",
        "eventsource",
        "eventsourcecode",
    ]
    _optional_properties = ["latitude", "longitude", "depth", "magnitude", "eventtime"]
    _optional_properties_fmt = {
        "latitude": "%.4f",
        "longitude": "%.4f",
        "depth": "%.1f",
        "magnitude": "%.1f",
        "eventtime": DATE_TIME_FMT,
    }

    def __init__(
        self,
        properties=None,
        local_files=None,
        local_directory=None,
        cancelfile=".cancel",
        product_properties=None,
        cmdline_args=None,
    ):
        """
        Create a PDLSender object using property settings, local files and
        directories to transfer/delete.

        Args:
            properties: Dictionary of properties that are needed for a PDL.
            local_files: List of local files which should be transferred to
                remote system.
            local_directory: Local directory which should be transferred to
                remote system.
            cancelfile: For Sender subclasses that actually send a file for
                cancel() actions, this allows the user to define what that
                file is called.
            product_properties: A dictionary of arbitrary properties that will
                be passed to PDL as --property-PROPERTY_NAME=PROPERTY_VALUE.
                These properties can be ints, floats, strings, or datetime
                objects.  Other kinds of objects will be converted to strings
                using the str() method - unpredictable results may follow.
            cmdline_args: A dictionary of additional command line arguments,
                which may have been added to the PDL client. There must
                not be leading "-" or "--" in front of the keys.

        """
        if product_properties is not None:
            self._product_properties = product_properties.copy()
        self.cmdline_args = None
        if cmdline_args is not None:
            self.cmdline_args = cmdline_args.copy()
        super().__init__(
            properties=properties,
            local_files=local_files,
            local_directory=local_directory,
            cancelfile=cancelfile,
        )

    def _replace_required_properties(self, cmd):
        for propkey in self._required_properties:
            propvalue = self._properties[propkey]
            cmd = cmd.replace("[" + propkey.upper() + "]", propvalue)
        return cmd

    def _replace_product_properties(self, cmd):
        if hasattr(self, "_product_properties"):
            prop_nuggets = []
            for propkey, propvalue in self._product_properties.items():
                if isinstance(propvalue, float):
                    prop_nuggets.append(f"--property-{propkey}={propvalue:.4f}")
                elif isinstance(propvalue, int):
                    prop_nuggets.append(f"--property-{propkey}={int(propvalue):d}")
                elif isinstance(propvalue, datetime.datetime):
                    pval_str = propvalue.strftime(DATE_TIME_FMT)[0:23]
                    prop_nuggets.append(f"--property-{propkey}={pval_str}")
                elif isinstance(propvalue, str):
                    prop_nuggets.append(f'--property-{propkey}="{propvalue}"')
                else:
                    prop_nuggets.append(f"--property-{propkey}={str(propvalue)}")
            cmd = cmd.replace("[PRODUCT_PROPERTIES]", " ".join(prop_nuggets))
        else:
            cmd = cmd.replace("[PRODUCT_PROPERTIES]", "")
        return cmd

    def _replace_files(self, cmd):
        if self._local_files:
            cmd = cmd.replace("[FILE]", f"--file={self._local_files[0]}")
        else:
            cmd = cmd.replace("[FILE]", "")
        if self._local_directory:
            cmd = cmd.replace("[DIRECTORY]", f"--directory={self._local_directory}")
        else:
            cmd = cmd.replace("[DIRECTORY]", "")

        return cmd

    def _replace_optional_properties(self, cmd):
        # add optional properties
        opts = set(self._optional_properties)
        props = set(self._properties.keys())
        hasopts = len(opts & props)

        if hasopts:
            opt_nuggets = []
            for propkey in self._optional_properties:
                if propkey in self._properties:
                    if propkey == "eventtime":
                        propr = self._properties[propkey]
                        prop_str = propr.strftime(DATE_TIME_FMT)[0:23]
                        opt_nuggets.append(f"--{propkey}={prop_str}Z")
                    else:
                        fmt = "--%s=" + self._optional_properties_fmt[propkey]
                        opt_nuggets.append(fmt % (propkey, self._properties[propkey]))
            cmd = cmd.replace("[OPTIONAL_PROPERTIES]", " ".join(opt_nuggets))
        else:
            cmd = cmd.replace("[OPTIONAL_PROPERTIES]", "")

        return cmd

    def _replace_cmdline_args(self, cmd):
        if self.cmdline_args is None:
            cmd = cmd.replace("[CMDLINE_ARGS]", " ")
            return cmd
        arg_nuggets = []
        for key, value in self.cmdline_args.items():
            nugget = f"--{key}={value}"
            arg_nuggets.append(nugget)
        cmd = cmd.replace("[CMDLINE_ARGS]", " ".join(arg_nuggets))
        return cmd

    def send(self):
        """Send local file or directory via PDL.

        Raises:
            Exception when:
             - number of local_files is greater than 1.
             - PDL command fails for any reason.

        Returns:
            Tuple containing number of files sent, and the standard output of
            the PDL command.
        """
        # we can really only support sending of one file, so error out
        # if someone has specified more than one.
        if len(self._local_files) > 1:
            raise ValueError("For PDL, you may only send one file at a time.")

        # build pdl command line from properties
        cmd = self._pdlcmd

        # make this an update status
        cmd = cmd.replace("[STATUS]", "UPDATE")

        # fill out the required properties
        cmd = self._replace_required_properties(cmd)

        # fill out any files or directories we'll be sending
        cmd = self._replace_files(cmd)

        # fill in all the product properties
        cmd = self._replace_product_properties(cmd)

        # fill in all the optional properties
        cmd = self._replace_optional_properties(cmd)

        # add any optional command line arguments used by PDL
        cmd = self._replace_cmdline_args(cmd)

        # call PDL on the command line
        retcode, stdout, stderr = get_command_output(cmd)
        if not retcode:
            ptype = self._properties["type"]
            fmt = f'Could not send product "{ptype}" due to error "{stdout + stderr}"'
            raise IOError(fmt)

        # return the number of files we just sent
        nfiles = 0
        if self._local_files:
            nfiles += 1
        if self._local_directory:
            nfiles += sum(
                [len(files) for r, d, files in os.walk(self._local_directory)]
            )
        numfiles = int(nfiles)
        error_msg = stdout.decode("utf-8")
        msg = (
            f'{numfiles:d} files sent successfully: resulting in output: "{error_msg}"'
        )
        return (nfiles, msg)

    def cancel(self, cancel_content=None):
        """Send a delete message out via PDL regarding the product in question.

        Args:
            cancel_content: String containing cancel message. This is NOT used
                in the implementation for this class.

        Returns:
            Standard output from PDL DELETE command.
        """
        # build pdl command line from properties
        self._properties["file"] = ""
        self._properties["directory"] = ""
        cmd = self._pdlcmd

        # make this a delete status
        cmd = cmd.replace("[STATUS]", "DELETE")

        # fill out the required properties
        cmd = self._replace_required_properties(cmd)

        # fill out any files or directories we'll be sending
        cmd = self._replace_files(cmd)

        # fill in all the product properties
        cmd = self._replace_product_properties(cmd)

        # fill in all the optional properties
        cmd = self._replace_optional_properties(cmd)

        retcode, stdout, stderr = get_command_output(cmd)
        if not retcode:
            ptype = self._properties["type"]
            fmt = (
                f'Could not delete product "{ptype}" due to error '
                f'"{stdout + stderr}"'
            )
            raise IOError(fmt)

        return stdout



#!/usr/bin/env python

# stdlib imports
import os.path
import shutil
import tempfile
from ftplib import FTP, error_perm

# local
from .sender import Sender


class FTPSender(Sender):
    """Class for sending and deleting files and directories via FTP.

    Required properties:
      - remote_host Name of FTP server.
      - remote_directory String path on remote_host where local files should
        be copied to.

    Optional properties:
      - user String user name, for FTP servers where anonymous login is not
        allowed.
      - password String password, for FTP servers where anonymous login is
        not allowed.

    Usage:
    sender = FTPSender(properties={'remote_host':'ftp.gov',
                                   'remote_directory':'/pub/incoming/event1'},
                       local_directory = '/home/user/event1')
    sender.send() => Creates remote url: ftp://ftp.gov/pub/incoming/event1 with contents of /home/user/event1 in it.

    OR

    sender = FTPSender(properties={'remote_host':'ftp.gov',
                                   'remote_directory':'/pub/incoming/event1'},
                       local_directory = '/home/user/event1/version1')
    sender.send() => Creates remote url: ftp://ftp.gov/pub/incoming/event1 with contents of /home/user/event1/version1 in it.

    OR
    sender = FTPSender(properties={'remote_host':'ftp.gov',
                                   'remote_directory':'/pub/incoming/event1'},
                       local_files = ['/home/user/event1/version1/file1.txt','/home/user/event1/version1/file2.txt'])
    sender.send() => Creates remote files: ftp://ftp.gov/pub/incoming/event1/file1.txt AND
                                           ftp://ftp.gov/pub/incoming/event1/file1.txt

    """

    _required_properties = ["remote_directory", "remote_host"]
    _optional_properties = ["user", "password"]

    def send(self):
        """
        Send any files or folders that have been passed to constructor.

        Returns:
            Tuple of Number of files sent to remote SSH server and message
            describing success.

        Raises:
            Exception when files cannot be sent to remote FTP server for any
            reason.
        """
        remote_host = self._properties["remote_host"]
        remote_folder = self._properties["remote_directory"]

        try:
            # this should put us at the top level folder
            ftp = self._setup()

            # send any files we want
            nfiles = 0
            for f in self._local_files:
                self.__sendfile(f, ftp)
                nfiles += 1

            # send everything in the directories we specified
            if self._local_directory is not None:
                local_directory = self._local_directory
                allfiles = self.getAllLocalFiles()
                for filename in allfiles:
                    try:
                        self._copy_file_with_path(
                            ftp, filename, remote_folder, local_folder=local_directory
                        )
                        nfiles += 1
                    except BaseException:
                        pass

            ftp.quit()
            return (
                nfiles,
                f"{int(nfiles):d} files were sent successfully to "
                f"{remote_host} {remote_folder}",
            )
        except Exception as obj:
            raise Exception(f'Could not send to {remote_host}.  Error "{str(obj)}"')

    def cancel(self):
        """
        Create a cancel file (named as indicated in constructor "cancelfile"
        parameter) in remote_directory on remote_host.

        Args:
            cancel_content: String containing text that should be written to
                the cancelfile.

        Returns:
            A string message describing what has occurred.
        """
        remote_host = self._properties["remote_host"]
        remote_folder = self._properties["remote_directory"]
        ftp = self._setup()

        # Create local .cancel file, then copy it to ftp server
        tempdir = tempfile.mkdtemp()
        try:
            tfile = os.path.join(tempdir, self._cancelfile)  # local file
            f = open(tfile, "wt")
            f.close()
            ftp.cwd(remote_folder)
            self.__sendfile(tfile, ftp)
        except Exception:
            raise Exception(
                f"Could not create .cancel file on {remote_host}/{remote_folder}"
            )
        finally:
            shutil.rmtree(tempdir)
        return (
            f"{self._cancelfile} file succesfully placed on "
            f"{remote_host} {remote_folder}"
        )

    def _setup(self):
        """Initiate an ftp connection with properties passed to constructor.

        Navigate to/create directory (as necessary) specified by
        remote_directory property.

        Returns:
            Instance of the ftplib.FTP class.
        """
        host = self._properties["remote_host"]
        remote_folder = self._properties["remote_directory"]
        # attempt to login to remote host
        try:
            ftp = FTP(host)
            if "user" in self._properties:
                user = self._properties["user"]
            else:
                user = ""
            if "password" in self._properties:
                password = self._properties["password"]
            else:
                password = ""
            if user == "":
                ftp.login()
            else:
                ftp.login(user, password)
        except error_perm:
            raise Exception(f"Could not login to remote host {host}")

        # attempt to cd to remote directory
        try:
            self._create_remote_directory(ftp, remote_folder)
        except Exception:
            ftp.quit()
            raise Exception(
                f'Could not navigate to directory "{remote_folder}" on '
                f"remote host {host}"
            )

        return ftp

    def _create_remote_directory(self, ftp, remote_directory):
        """Create directory (recursively) on remote_host.

        Args:
            ftp: ftplib.FTP instance.
            remote_directory: String path of directory on remote system which
                needs to be created.

        Raises:
            Exception when unable to create remote_directory.
        """
        # attempt to cd to remote directory
        ftp.cwd("/")
        try:
            ftp.cwd(remote_directory)
        except error_perm:
            dirparts = self._split(remote_directory)
            for directory in dirparts:
                try:
                    ftp.cwd(directory)
                except error_perm:
                    try:
                        ftp.mkd(directory)
                        ftp.cwd(directory)
                    except error_perm:
                        raise Exception(f"Unable to create subdirectory {directory}.")

    def _copy_file_with_path(self, ftp, local_file, remote_folder, local_folder=None):
        """
        Copy local_file to remote_folder, preserving relative path and creating
        required sub-directories.

        Usage:
         local_file: /home/user/data/events/us2016abcd/data_files/datafile.txt
         remote_folder: /data/archive/events
         local_folder: /home/user/data/events/us2016abcd

         would create:
           /data/archive/events/us2016abcd/data_files/datafile.txt

        local_file: /home/user/data/events/us2016abcd/data_files/datafile.txt
        remote_folder: /data/archive/events/us2016abcd
        local_folder: None

        would create:
          /data/archive/events/us2016abcd/datafile.txt

        Args:
            local_file: Local file to copy.
            remote_folder: Remote folder to copy local files to.
            local_folder: Top of local directory where file copying started.
                If None, local_file should be copied to a file of the same
                name (not preserving path) into remote_folder.
        """
        if local_folder is None:
            ftp.cwd(remote_folder)
            self.__sendfile(local_file, ftp)
        else:
            local_parts = (
                local_file.replace(local_folder, "")
                .strip(os.path.sep)
                .split(os.path.sep)
            )
            remote_parts = self._split(remote_folder)
            all_parts = remote_parts + local_parts
            remote_file = "/" + "/".join(all_parts)
            print(remote_file)
            remfolder, _ = self._path_split(remote_file)
            try:
                ftp.cwd(remfolder)
            except error_perm:
                self._create_remote_directory(ftp, remfolder)

            self.__sendfile(local_file, ftp)
            ftp.cwd(remote_folder)

    def __sendfile(self, filename, ftp):
        """Internal function used to send a file using an FTP object.

        Args:
            filename: Local filename
            ftp: Instance of FTP object.
        """
        # in case somebody is polling for this file,
        # make a temporary file first, then rename it
        # so the poller doesn't grab it before its finished transferring.
        _, fpath = os.path.split(filename)  # this is a local file
        tmpfile = fpath + ".tmp"
        cmd = "STOR " + tmpfile
        # we don't tell the ftp server about the local path to the file

        # actually send the file
        ftp.storbinary(cmd, open(filename, "rb"), 1024)
        # rename it to the desired destination
        ftp.rename(tmpfile, fpath)

    def _join(self, *path_parts):
        return "/" + "/".join(path_parts)

    def _split(self, path):
        return path.strip("/").split("/")

    def _path_split(self, path):
        parts = path.strip("/").split("/")
        fname = parts[-1]
        fpath = "/" + "/".join(parts[0:-1])
        return (fpath, fname)



#!/usr/bin/env python

# stdlib imports
import shutil
import os.path
import datetime

# local imports
from .sender import Sender


class CopySender(Sender):
    """
    Class for sending and deleting files and directories via system copy
    and delete.

    CopySender creates a remote directory if one does not already exist, and
    copies files and directories to that remote directory.  Files are simply
    copied to the remote_directory.  Files in the local directory are copied
    to the remote directory, and sub-directories under the local directory are
    preserved in the remote directory.

    For example:
      If the local_directory is /home/user/event1/, which contains:
      /home/user/event1/datafile1.txt
      /home/user/event1/other/datafile2.txt

      and the remote_directory is /data/event1, then it will contain:
      /data/event1/datafile1.txt
      /data/event1/other/datafile2.txt

    The cancel() method implemented in this class behaves as follows:
    cs = CopySender(properties={'remote_directory':'/data/event1'},
                    local_directory='/home/user/event1',cancelfile='CANCEL')
    #Sending...
    cs.cancel() #=>This creates a file called /data/event1/CANCEL.

    Required properties:
      - remote_directory String indicating which directory input content
        should be copied to.

    """

    _required_properties = ["remote_directory"]
    _optional_properties = []

    def send(self):
        """Send any files or folders that have been passed to constructor.

        This method deletes any previous cancel files that may exist in the
        remote_directory.

        Returns:
            Tuple containing number of files sent to local directory, and a
            message describing success.

        """
        # copy directories to remote location, changing remote name to desired
        # alias
        remote_folder = self._properties["remote_directory"]

        # remove any previous cancel files sent to remote_folder
        cancelfile = os.path.join(remote_folder, self._cancelfile)
        if os.path.isfile(cancelfile):
            os.remove(cancelfile)

        nfiles = 0
        if not os.path.isdir(remote_folder):
            os.makedirs(remote_folder)
        if self._local_directory:
            local_folder = self._local_directory
            # recursively find all of the files locally, then
            # copy them to corresponding remote folder structure
            allfiles = self.getAllLocalFiles()
            for filename in allfiles:
                self._copy_file_with_path(
                    filename, remote_folder, local_folder=local_folder
                )
                nfiles += 1

            nfiles += sum([len(files) for r, d, files in os.walk(local_folder)])

        # copy files to remote location
        for filename in self._local_files:
            self._copy_file_with_path(filename, remote_folder)
            nfiles += 1

        return (nfiles, f"{int(nfiles):d} files sent successfully using CopySender.")

    def cancel(self, cancel_content=None):
        """
        Create a cancel file (named as indicated in constructor "cancelfile"
        parameter) in remote_directory.

        Args:
            cancel_content: String containing text that should be written to
                the cancelfile.

        Returns:
            A string message describing what has occurred.
        """
        remote_folder = self._properties["remote_directory"]
        cancelfile = os.path.join(remote_folder, self._cancelfile)
        f = open(cancelfile, "wt")
        if cancel_content is not None:
            f.write(cancel_content)
        f.close()
        return f"A .cancel file has been placed in remote directory {remote_folder}."

    def _copy_file_with_path(self, local_file, remote_folder, local_folder=None):
        """
        Copy local_file to remote_folder, preserving relative path and creating
        required sub-directories.

        Usage:
         local_file: /home/user/data/events/us2016abcd/data_files/datafile.txt
         remote_folder: /data/archive/events
         local_folder: /home/user/data/events/us2016abcd

         would create:
           /data/archive/events/us2016abcd/data_files/datafile.txt

        local_file: /home/user/data/events/us2016abcd/data_files/datafile.txt
        remote_folder: /data/archive/events/us2016abcd
        local_folder: None

        would create:
          /data/archive/events/us2016abcd/datafile.txt

        Args:
            local_file: Local file to copy.
            remote_folder: Remote folder to copy local files to.
            local_folder: Top of local directory where file copying started.
                If None, local_file should be copied to a file of the same
                name (not preserving path) into remote_folder.
        """
        if local_folder is not None:
            local_parts = (
                local_file.replace(local_folder, "")
                .strip(os.path.sep)
                .split(os.path.sep)
            )
            remote_parts = remote_folder.strip(os.path.sep).split(os.path.sep)
            all_parts = [os.path.sep] + remote_parts + local_parts
            remote_file = os.path.join(*all_parts)
            root, rfile = os.path.split(remote_file)
            if not os.path.isdir(root):
                os.makedirs(root)
        else:
            root, tfile = os.path.split(local_file)
            remote_file = os.path.join(remote_folder, tfile)
        remote_tmp_file = (
            remote_file + ".tmp_" + datetime.datetime.utcnow().strftime("%Y%m%d%H%M%S")
        )
        shutil.copyfile(local_file, remote_tmp_file)
        os.rename(remote_tmp_file, remote_file)



# stdlib imports
import email.utils
import mimetypes
import os.path
import shutil
import ssl
import tempfile
import zipfile
from email import encoders
from email.message import Message
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from smtplib import (
    SMTP,
    SMTP_SSL,
    SMTPDataError,
    SMTPHeloError,
    SMTPRecipientsRefused,
    SMTPSenderRefused,
)

# local
from .sender import Sender

# default maximum number of bcc email addresses per message.
# Not sure what this is normally set to on typical SMTP server,
# so being conservative here
MAX_BCC = 25
DEFAULT_CANCEL_MSG = "This is a cancel message."

SSL_PORT = 465
STARTTLS_PORT = 587
TLS_PORT = 25


def get_server(method, smtp_server):
    if method == "SSL":
        port = SSL_PORT
        context = ssl.create_default_context()
        server = SMTP_SSL(smtp_server, port, context=context)
    elif method == "STARTTLS":
        port = STARTTLS_PORT
        context = ssl.create_default_context()
        server = SMTP(smtp_server, port)
        server.starttls(context=context)
    elif method == "TLS":
        port = TLS_PORT
        context = ssl.create_default_context()
        server = SMTP(smtp_server, port)
        server.starttls(context=context)
    else:
        server = SMTP(smtp_server)
    return server


class EmailSender(Sender):
    """Class for sending files via email to any number of recipients.

    EmailSender will send any number (within email server file size attachment
    limits) of files and/or directories to a set of recipients.  The file
    attachments can be zipped together into a single .zip  file attachment.
    If you are sending more than one file, creating a zip file is HIGHLY
    recommended.

    By default, the emails will be sent using Bcc lists of 25, meaning that a
    email messages to multiple recipients will be sent to one person, with 25
    Bcc recipients.  This value can be  changed with the optional max_bcc
    property.  If you do not happen to control the SMTP server through which
    you are sending emails, or know what its settings are, it is best (in the
    author's opinion) to be conservative in setting this value.

    To send a file to a list of email recipients:
    props = {}
    props['recipients'] = ['fred@foo.com','barney@bar.org']
    props['smtp_servers'] = [my.smtp.server.org]
    props['sender'] = 'mrslate@quarry.org'
    props['subject'] = "You're fired!"
    props['message'] = 'You two idiots are thicker than the rocks you break!'

    sender = EmailSender(properties=props,local_files=['/home/mrslate/hammer.gif'])
    sender.send()

    To send a cancel message to the same list of email recipients:
    sender.cancel(cancel_content='Never mind - I like the way you guys work.')

    To send a zipped directory to the same list of email recipients:
    props = {}
    props['recipients'] = ['fred@foo.com','barney@bar.org']
    props['smtp_servers'] = [my.smtp.server.org]
    props['sender'] = 'mrslate@quarry.org'
    props['subject'] = "You're fired!"
    props['message'] = 'You two idiots are thicker than the rocks you break!'
    props['zip_file'] = 'allfiles.zip'
    sender = EmailSender(properties=props,local_directory='/home/mrslate/termination_docs'])
    sender.send()

    To send individual email messages (not using Bcc functionality) to a
    number of recipients:
    props = {}
    props['recipients'] = ['fred@foo.com','barney@bar.org']
    props['smtp_servers'] = [my.smtp.server.org]
    props['sender'] = 'mrslate@quarry.org'
    props['subject'] = "You're fired!"
    props['message'] = 'You two idiots are thicker than the rocks you break!'
    props['max_bcc'] = 0

    To send batch email messages (using Bcc functionality), preserving
    anonymity betweeen recipients by using the sender as the primary recipient:
    props = {}
    props['recipients'] = ['fred@foo.com','barney@bar.org']
    props['smtp_servers'] = [my.smtp.server.org]
    props['sender'] = 'mrslate@quarry.org'
    props['subject'] = "You're fired!"
    props['message'] = 'You two idiots are thicker than the rocks you break!'
    props['max_bcc'] = 25

    To send batch email messages (using Bcc functionality), preserving
    anonymity betweeen recipients by setting the primary_recipient property:
    props = {}
    props['recipients'] = ['fred@foo.com','barney@bar.org']
    props['smtp_servers'] = [my.smtp.server.org]
    props['sender'] = 'mrslate@quarry.org'
    props['subject'] = "You're fired!"
    props['message'] = 'You two idiots are thicker than the rocks you break!'
    props['primary_recipient'] = 'employees@quarry.org'
    props['max_bcc'] = 25

    Required properties:
      - smtp_servers List of strings indicating hostnames for SMTP servers to
        which you have permissions to connect.
      - sender Email address which will appear in the From: field in the
        recipient's email.
      - subject String containing the email subject line.
      - recipients List of valid email addresses to which message will be sent.
      - message Message which will be sent.
    Optional properties:
      - zip_file Zip file containing content to be attached to message
      - max_bcc Maximum number of addresses that can be added to a Bcc: list.
                Defaults to 25.
      - protocol Desired encryption protocol. One of: "TLS", "STARTTLS", or "SSL".
    """

    _required_properties = [
        "smtp_servers",
        "sender",
        "subject",
        "recipients",
        "message",
    ]

    _optional_properties = ["zip_file", "max_bcc", "port", "protocol"]

    def send(self):
        """
        Send a message to intended recipients with or without attachment.

        This method should determine the MIME type of the attachment and
        insert it into the message, along with the specified text.

        Raises:
          Exception when:
           - Attachment is not a valid file.
           - There is one of a number of errors connecting to email servers.
        """
        tempdir = tempfile.mkdtemp()
        zip_file = None
        if "zip_file" in self._properties:
            zip_file = self._properties["zip_file"]
        max_bcc = MAX_BCC
        if "max_bcc" in self._properties:
            max_bcc = self._properties["max_bcc"]

        # if using Bcc, try to maintain privacy between recipients by using
        # either an empty string (default) or use the primary_recipient
        # property, if set.
        primary_recipient = None
        if max_bcc:
            if "primary_recipient" in self._properties:
                primary_recipient = self._properties["primary_recipient"]
            else:
                primary_recipient = ""

        sender = self._properties["sender"]
        subject = self._properties["subject"]
        text = self._properties["message"]
        smtp_servers = self._properties["smtp_servers"]
        protocol = self._properties.get("protocol", "default")
        # send email to all recipients, attaching files as necessary or zipping
        # into one file to be attached.
        try:
            address_tuples = _split_addresses(
                self._properties["recipients"], max_bcc, primary_recipient
            )

            for address, bcc in address_tuples:
                attachments = []
                if len(self._local_files) or self._local_directory:
                    if zip_file is not None:
                        # create a zip file with all of the contents
                        zfilename = os.path.join(tempdir, zip_file)
                        myzip = zipfile.ZipFile(
                            zfilename, "w", compression=zipfile.ZIP_DEFLATED
                        )
                        for filename in self._local_files:
                            root, arcname = os.path.split(filename)
                            myzip.write(filename, arcname)
                        if self._local_directory:
                            allfiles = self.getAllLocalFiles()
                            for filename in allfiles:
                                arcname = filename.replace(self._local_directory, "")
                                myzip.write(filename, arcname)
                        myzip.close()
                        attachments.append(zfilename)
                    else:
                        for filename in self._local_files:
                            attachments.append(filename)

                            all_files = self.getAllLocalFiles()
                        for filename in all_files:
                            attachments.append(filename)

                if not len(attachments):
                    msg = MIMEText(text)
                    msg["From"] = sender
                    msg["To"] = address
                    msg["Subject"] = subject
                    msg["Date"] = email.utils.formatdate()
                    if bcc is not None:
                        bccstr = ", ".join(bcc)
                        msg["Bcc"] = bccstr
                    msgtxt = msg.as_string()
                else:
                    msgtxt = _get_encoded_message(
                        address, subject, text, sender, attachments, bcc=bcc
                    )
                if bcc is not None:
                    all_addresses = [address] + bcc
                else:
                    all_addresses = [address]
                _send_email(sender, all_addresses, msgtxt, smtp_servers, protocol)
        except Exception as e:
            raise Exception(
                f'Could not send mail to {address} with EmailSender. "{str(e)}"'
            )
        finally:
            shutil.rmtree(tempdir)
        nfiles = len(self._local_files)
        if self._local_directory is not None:
            nfiles += sum(
                [len(files) for r, d, files in os.walk(self._local_directory)]
            )
        num_files = int(nfiles)
        num_recipients = int(len(self._properties["recipients"]))
        fmt = f"{num_files:d} files successfully sent to {num_recipients:d} recipients."
        return (nfiles, fmt)

    def cancel(self, cancel_content=None):
        """Send a cancel message to list of recipients.

        Args:
            cancel_content: String containing text that should be sent to
                recipients (default 'This is a cancel message.')

        Returns:
            A string message describing what has occurred.
        """
        # send a cancel message to all recipients
        sender = self._properties["sender"]
        protocol = self._properties.get("protocol", "default")
        if "cancel_subject" in self._properties:
            subject = self._properties["cancel_subject"]
        else:
            subject = self._properties["subject"]
        if "cancel_content" in self._properties:
            cancel_content = self._properties["cancel_content"]

        max_bcc = MAX_BCC
        if "max_bcc" in self._properties:
            max_bcc = self._properties["max_bcc"]

        # if using Bcc, try to maintain privacy between recipients by using
        # either an empty string (default) or use the primary_recipient
        # property, if set.
        primary_recipient = None
        if max_bcc:
            if "primary_recipient" in self._properties:
                primary_recipient = self._properties["primary_recipient"]
            else:
                primary_recipient = ""

        address_tuples = _split_addresses(
            self._properties["recipients"], max_bcc, primary_recipient
        )
        smtp_servers = self._properties["smtp_servers"]
        if cancel_content is None:
            cancel_content = DEFAULT_CANCEL_MSG
        for address, bcc in address_tuples:
            msg = MIMEText(cancel_content)
            msg["From"] = sender
            msg["To"] = address
            msg["Subject"] = subject
            msg["Date"] = email.utils.formatdate()
            bccstr = ", ".join(bcc)
            msg["Bcc"] = bccstr
            msgtxt = msg.as_string()
            all_addresses = [address] + bcc
            _send_email(sender, all_addresses, msgtxt, smtp_servers, protocol)


def _split_addresses(recipients, max_bcc, primary_recipient):
    """Split addresses into a list of tuples of (recipient,bcclist).

    Args:
        recipients: List of intended email recipients.

    Returns:
        List of tuples, where each tuple consists of:
        - Email recipient.
        - list of recipients who will be bcc'd on the email to the first
          recipient.
    """

    tuples = []
    if max_bcc == 0:
        for recipient in recipients:
            tuples.append((recipient, []))
        return tuples
    istart = 0
    iend = max_bcc
    while istart < len(recipients):
        address = primary_recipient
        iend = min(len(recipients), istart + (max_bcc - 1))
        bcclist = recipients[istart:iend]
        tuples.append((address, bcclist))
        istart = iend

    return tuples


def _send_email(sender, address, msgtxt, smtp_servers, protocol):
    """Send email to given address(es).

    Args:
        sender: From email address.
        address: Single email address or list of addresses.
        msgtxt: MIMEText email message encoded as a string.
        smtp_servers: List of smtp servers to try to send from.

    Raises:
        Exception if sending the message failed for any reason.
    """
    messageSent = False
    errormsg = []
    # let's try all of the email servers we know about before
    # admitting defeat...
    servername = ""
    for server in smtp_servers:
        # print 'Trying server %s' % (server)
        try:
            session = get_server(protocol, server)
            resp, servername = session.helo()
            session.sendmail(sender, address, msgtxt)
            messageSent = True
            session.quit()
            break
        except SMTPRecipientsRefused:
            errormsg.append({server: "Recipients refused"})
            continue
        except SMTPHeloError:
            errormsg.append({server: "Server did not respond to hello"})
            continue
        except SMTPSenderRefused:
            errormsg.append({server: "Server refused sender address"})
            continue
        except SMTPDataError:
            errormsg.append({server: "Server responded with an unexpected error code"})
            continue
        except BaseException:
            errormsg.append({server: "Connection to server failed (possible timeout)"})

    if not messageSent:
        errstr = (
            f"The message to {address} was not sent. "
            "The server error messages are below:"
        )
        for errdict in errormsg:
            errstr = errstr + str(errdict)
        raise Exception(str(errstr))

    print(f'Message sent to "{address}" via SMTP server "{servername}"')
    # if bcc is not None:
    #     print('Bcc: %s' % ','.join(bcc))


def _get_encoded_message(address, subject, text, sender, attachments, bcc=None):
    """Private method for encoding attachment into a MIME string.

    Args:
        address: Primary (not Bcc) email recipient.
        subject: String email subject line.
        text: String text of email message.
        attachments: List of files to be attached to message.
        bcc: List of email addresses to Bcc.

    Returns:
        Message text.
    """
    outer = MIMEMultipart()
    outer["Subject"] = subject
    outer["To"] = address
    outer["From"] = sender
    outer["Date"] = email.utils.formatdate()
    if bcc is not None:
        outer["Bcc"] = ", ".join(bcc)

    # insert the text into the email as a MIMEText part...
    firstSubMsg = Message()
    firstSubMsg["Content-type"] = "text/plain"
    firstSubMsg["Content-transfer-encoding"] = "7bit"
    firstSubMsg.set_payload(text)
    outer.attach(firstSubMsg)

    for attachment in attachments:
        ctype, encoding = mimetypes.guess_type(attachment)
        msg = None
        if ctype is None or encoding is not None:
            # No guess could be made, or the file is encoded (compressed), so
            # use a generic bag-of-bits type.
            ctype = "application/octet-stream"
        maintype, subtype = ctype.split("/", 1)
        if maintype == "text":
            fp = open(attachment)
            # Note: we should handle calculating the charset
            msg = MIMEText(fp.read(), _subtype=subtype)
            fp.close()
        elif maintype == "image":
            fp = open(attachment, "rb")
            msg = MIMEImage(fp.read(), _subtype=subtype)
            fp.close()
        elif maintype == "audio":
            fp = open(attachment, "rb")
            msg = MIMEAudio(fp.read(), _subtype=subtype)
            fp.close()
        elif maintype == "application":
            fp = open(attachment, "rb")
            msg = MIMEApplication(fp.read(), _subtype=subtype)
            fp.close()
        else:
            fp = open(attachment, "rb")
            msg = MIMEBase(maintype, subtype)
            msg.set_payload(fp.read())
            fp.close()
            # Encode the payload using Base64
            encoders.encode_base64(msg)

        msg.add_header(
            "Content-Disposition", "attachment", filename=os.path.basename(attachment)
        )
        outer.attach(msg)

    return outer.as_string()



#!/usr/bin/env python

# stdlib imports
import os.path
import datetime
import getpass

# depends on https://github.com/jbardin/scp.py
# pip install git+git://github.com/jbardin/scp.py.git
from paramiko import SSHClient
from scp import SCPClient
from .sender import Sender

DEFAULT_PORT = 22


class SecureSender(Sender):
    """Class for sending files to a remote Unix-like system using ssh.

    SecureSender creates a remote directory if one does not already exist, and
    copies files and directories to that remote directory.  Files are simply
    copied to the remote_directory.  Files in the local directory are copied
    to the remote directory, and sub-directories under the local directory are
    preserved in the remote directory.

    For example:
      If the local_directory is /home/user/event1/, which contains:
      /home/user/event1/datafile1.txt
      /home/user/event1/other/datafile2.txt

      and the remote_directory is /data/event1, then it will contain:
      /data/event1/datafile1.txt
      /data/event1/other/datafile2.txt

    The cancel() method implemented in this class behaves as follows:
    cs = CopySender(properties={'remote_directory':'/data/event1'},
                    local_directory='/home/user/event1',cancelfile='CANCEL')
    #Sending...
    cs.cancel() #=>This creates a file called /data/event1/CANCEL.

    Required properties:
      - remote_host String indicating the name of the remote host to which
        files should be copied.
      - remote_directory String indicating which directory input content
        should be copied to.
      - private_key Path to local RSA/DSA private key file which has been
        configured on the remote system.
    Optional properties:
      - remote_port: This defaults to 22, set to whatever port your
                     remote system has opened up for SSH access.
    """

    _required_properties = ["private_key", "remote_host", "remote_directory"]
    _optional_properties = []

    def send(self):
        """
        Send any files or folders that have been passed to constructor to a
        remote Unix-like system.

        Returns:
            Tuple of Number of files sent to remote SSH server and a message
            describing success.

        Raises:
            Exception when unable to create remote folder on remote host, or
            copy file to remote host.
        """
        # figure out the remote folder
        remote_folder = self._properties["remote_directory"]
        remote_host = self._properties["remote_host"]
        # connect to remote system and instantiate scp object
        nfiles = 0
        ssh = self._connect()
        # SCPCLient takes a paramiko transport as its only argument
        scp = SCPClient(ssh.get_transport())

        # make sure the remote system has directory or it can be created.
        res = self._make_remote_folder(scp, ssh, remote_folder)
        if not res:
            msg = (
                f"Unable to create remote folder {remote_folder} "
                f"on host {remote_host}"
            )
            raise Exception(msg)

        # do the copying
        try:
            allfiles = []
            if len(self._local_files):
                allfiles = self._local_files
                for filename in allfiles:
                    self._copy_file_with_path(scp, ssh, filename, remote_folder)

            if self._local_directory is not None:
                allfiles = self.getAllLocalFiles()
                for filename in allfiles:
                    self._copy_file_with_path(
                        scp,
                        ssh,
                        filename,
                        remote_folder,
                        local_folder=self._local_directory,
                    )

        except Exception as obj:
            rhost = self._properties["remote_host"]
            raise Exception(
                f'Could not send files to remote host {rhost}: "{str(obj)}"'
            )

        nfiles += len(self._local_files)
        if self._local_directory is not None:
            nfiles += sum(
                [len(files) for r, d, files in os.walk(self._local_directory)]
            )
        scp.close()
        ssh.close()
        msg = f"{int(nfiles):d} files sent to remote host {remote_host}"
        return (nfiles, msg)

    def cancel(self, cancel_content=None):
        """
        Create a cancel file (named as indicated in constructor "cancelfile"
        parameter) in remote_directory on remote_host.

        Args:
            cancel_content: String containing text that should be written to
                the cancelfile.

        Returns:
            A string message describing what has occurred.
        """
        remote_folder = self._properties["remote_directory"]
        remote_host = self._properties["remote_host"]
        ssh = self._connect()
        cancelfile = os.path.join(remote_folder, self._cancelfile)
        echo_cmd = f"echo {cancel_content} > {cancelfile}"
        _, stdout, stderr1 = ssh.exec_command(echo_cmd)
        chk_cmd = f"[ -e {cancelfile} ];echo $?"
        _, stdout, _ = ssh.exec_command(chk_cmd)
        exists = not int(stdout.read().decode("utf-8").strip())
        if not exists:
            err = stderr1.read().decode("utf-8").strip()
            fmt = (
                f"Could not create {cancelfile} file on "
                f"remote_host {remote_host} due to error: {err}"
            )
            raise Exception(fmt)

        msg = "A .cancel file has been placed in remote " f"directory {remote_folder}."
        return msg

    def _connect(self):
        """Initiate an ssh connection with properties passed to constructor.

        Returns:
            Instance of the paramiko SSHClient class.

        Raises:
            Exception if connection to remote host fails.
        """
        ssh = SSHClient()
        # load hosts found in ~/.ssh/known_hosts
        # should we not assume that the user has these configured already?
        ssh.load_system_host_keys()
        remote_user = getpass.getuser()
        if "remote_user" in self._properties:
            remote_user = self._properties["remote_user"]
        remote_port = DEFAULT_PORT
        if "remote_port" in self._properties:
            remote_port = self._properties["remote_port"]
        try:
            ssh.connect(
                self._properties["remote_host"],
                port=remote_port,
                username=remote_user,
                key_filename=self._properties["private_key"],
                compress=True,
            )
        except Exception as obj:
            msg = (
                "Could not connect with private key "
                f"file {self._properties['private_key']}: "
                f"Error '{str(obj)}"
            )
            raise Exception(msg)
        return ssh

    def _copy_file_with_path(
        self, scp, ssh, local_file, remote_folder, local_folder=None
    ):
        """
        Copy local_file to remote_folder, preserving relative path and creating
        required sub-directories.

        Usage:
         local_file: /home/user/data/events/us2016abcd/data_files/datafile.txt
         remote_folder: /data/archive/events
         local_folder: /home/user/data/events/us2016abcd

         would create:
           /data/archive/events/us2016abcd/data_files/datafile.txt

        local_file: /home/user/data/events/us2016abcd/data_files/datafile.txt
        remote_folder: /data/archive/events/us2016abcd
        local_folder: None

        would create:
          /data/archive/events/us2016abcd/datafile.txt


        Args:
            scp: SCPClient instance.
            ssh: SSHClient instance.
            local_file: Local file to copy.
            remote_folder: Remote folder to copy local files to.
            local_folder: Top of local directory where file copying started.
                If None, local_file should be copied to a file of the same
                name (not preserving path) into remote_folder.
        """
        remote_host = self._properties["remote_host"]
        if local_folder is not None:
            local_parts = (
                local_file.replace(local_folder, "")
                .strip(os.path.sep)
                .split(os.path.sep)
            )
            remote_parts = remote_folder.strip(os.path.sep).split(os.path.sep)
            all_parts = [os.path.sep] + remote_parts + local_parts
            remote_file = os.path.join(*all_parts)
            root, _ = os.path.split(remote_file)
            _, isdir = self._check_remote_folder(ssh, root)
            if not isdir:
                res = self._make_remote_folder(scp, ssh, root)
                if not res:
                    fmt = (
                        f"Could not copy local file {local_file} "
                        f"to folder {remote_folder} "
                        f"on host {remote_host}"
                    )
                    raise Exception(fmt)

        else:
            root, tfile = os.path.split(local_file)
            remote_file = os.path.join(remote_folder, tfile)
        remote_tmp_file = (
            remote_file + ".tmp_" + datetime.datetime.utcnow().strftime("%Y%m%d%H%M%S")
        )
        print(f"Copying {local_file}...")
        scp.put(local_file, remote_tmp_file)
        rename_cmd = f"mv {remote_tmp_file} {remote_file}"
        ssh.exec_command(rename_cmd)

    def _check_remote_folder(self, ssh, remote_folder):
        """Check to see if remote folder exists and is a directory.

        Args:
            ssh: SSHClient instance.
            remote_folder: Remote folder to copy local files to.

        Returns:
            Tuple with two booleans -- (does a file or directory of this name
            exist, is it a directory?)
        """
        chk_cmd1 = f"[ -e {remote_folder} ];echo $?"
        _, stdout, _ = ssh.exec_command(chk_cmd1)
        exists = not int(stdout.read().decode("utf-8").strip())
        chk_cmd2 = f"[ -d {remote_folder} ];echo $?"
        _, stdout, _ = ssh.exec_command(chk_cmd2)
        isdir = not int(stdout.read().decode("utf-8").strip())
        return (exists, isdir)

    def _make_remote_folder(self, scp, ssh, remote_folder):
        """Make a folder on remote system.

        Args:
            scp: SCPClient instance.
            ssh: SSHClient instance.
            remote_folder: Remote folder to copy local files to.

        Returns:
            Boolean indicating success or failure.
        """
        exists, isdir = self._check_remote_folder(ssh, remote_folder)
        chk_cmd1 = f"[ -d {remote_folder} ];echo $?"
        if not isdir:
            if exists:
                rm_cmd = f"rm {remote_folder}"
                _, stdout, _ = ssh.exec_command(rm_cmd)
                _, stdout, _ = ssh.exec_command(chk_cmd1)
                exists = not int(stdout.read().decode("utf-8").strip())
            if not exists:
                mk_cmd = f"mkdir -p {remote_folder}"
                _, stdout, _ = ssh.exec_command(mk_cmd)
                exists, isdir = self._check_remote_folder(ssh, remote_folder)
                if not isdir:
                    return False
        return True
